/*
**
**		AmiTCP Dumb Nameserver
**      ----------------------
**		Amiga version hacked & slain by Oliver Wagner (o.wagner@pluribus.wupper.de)
**
**		very loosely based on JoranNameserver for AmiTCP 2.x
**
**		compile with SAS/C 6.55
**
*/

#include <bsdsocket.h>
#include <syslog.h>

#define BYTE_ORDER BIG_ENDIAN
#include <arpa/nameser.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

// Amiga stuff
#include <proto/dos.h>
#include <dos/stdio.h>
#include <proto/utility.h>

#include <string.h>

#include "rev.h"
char version[] = {"$VER: AmiTCP-NameServer " VERTAG };

long __oslibversion = 37;

/*
 * ++Copyright++ 1985
 * -
 * Copyright (c) 1985 Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 	This product includes software developed by the University of
 * 	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * -
 * Portions Copyright (c) 1993 by Digital Equipment Corporation.
 * 
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies, and that
 * the name of Digital Equipment Corporation not be used in advertising or
 * publicity pertaining to distribution of the document or software without
 * specific, written prior permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND DIGITAL EQUIPMENT CORP. DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL DIGITAL EQUIPMENT
 * CORPORATION BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * -
 * --Copyright--
 */

/*
 * Expand compressed domain name 'comp_dn' to full domain name.
 * 'msg' is a pointer to the begining of the message,
 * 'eomorig' points to the first location after the message,
 * 'exp_dn' is a pointer to a buffer of size 'length' for the result.
 * Return size of compressed name or -1 if there was an error.
 */
static dn_expand(const u_char *msg, const u_char *eomorig, const u_char *comp_dn, u_char *exp_dn, int length)
{
	 u_char *cp, *dn;
	 int n, c;
	u_char *eom;
	int len = -1, checked = 0;

	dn = exp_dn;
	cp = (u_char *)comp_dn;
	eom = exp_dn + length;
	/*
	 * fetch next label in domain name
	 */
	while (n = *cp++) {
		/*
		 * Check for indirection
		 */
		switch (n & INDIR_MASK) {
		case 0:
			if (dn != exp_dn) {
				if (dn >= eom)
					return (-1);
				*dn++ = '.';
			}
			if (dn+n >= eom)
				return (-1);
			checked += n + 1;
			while (--n >= 0) {
				if ((c = *cp++) == '.') {
					if (dn + n + 2 >= eom)
						return (-1);
					*dn++ = '\\';
				}
				*dn++ = c;
				if (cp >= eomorig)	/* out of range */
					return(-1);
			}
			break;

		case INDIR_MASK:
			if (len < 0)
				len = cp - comp_dn + 1;
			cp = (u_char *)msg + (((n & 0x3f) << 8) | (*cp & 0xff));
			if (cp < msg || cp >= eomorig)	/* out of range */
				return(-1);
			checked += 2;
			/*
			 * Check for loops in the compressed name;
			 * if we've looked at the whole message,
			 * there must be a loop.
			 */
			if (checked >= eomorig - msg)
				return (-1);
			break;

		default:
			return (-1);			/* flag error */
		}
	}
	*dn = '\0';
	if (len < 0)
		len = cp - comp_dn;
	return (len);
}

/*
 * Search for expanded name from a list of previously compressed names.
 * Return the offset from msg if found or -1.
 * dnptrs is the pointer to the first name on the list,
 * not the pointer to the start of the message.
 */
static int dn_find(u_char *exp_dn, u_char *msg, u_char **dnptrs, u_char **lastdnptr)
{
	 u_char *dn, *cp, **cpp;
	 int n;
	u_char *sp;

	for (cpp = dnptrs; cpp < lastdnptr; cpp++) {
		dn = exp_dn;
		sp = cp = *cpp;
		while (n = *cp++) {
			/*
			 * check for indirection
			 */
			switch (n & INDIR_MASK) {
			case 0:		/* normal case, n == len */
				while (--n >= 0) {
					if (*dn == '.')
						goto next;
					if (*dn == '\\')
						dn++;
					if (*dn++ != *cp++)
						goto next;
				}
				if ((n = *dn++) == '\0' && *cp == '\0')
					return (sp - msg);
				if (n == '.')
					continue;
				goto next;

			default:	/* illegal type */
				return (-1);

			case INDIR_MASK:	/* indirection */
				cp = msg + (((n & 0x3f) << 8) | *cp);
			}
		}
		if (*dn == '\0')
			return (sp - msg);
	next:	;
	}
	return (-1);
}

/*
 * Compress domain name 'exp_dn' into 'comp_dn'.
 * Return the size of the compressed name or -1.
 * 'length' is the size of the array pointed to by 'comp_dn'.
 * 'dnptrs' is a list of pointers to previous compressed names. dnptrs[0]
 * is a pointer to the beginning of the message. The list ends with NULL.
 * 'lastdnptr' is a pointer to the end of the arrary pointed to
 * by 'dnptrs'. Side effect is to update the list of pointers for
 * labels inserted into the message as we compress the name.
 * If 'dnptr' is NULL, we don't try to compress names. If 'lastdnptr'
 * is NULL, we don't update the list.
 */
static dn_comp( const u_char *exp_dn, u_char *comp_dn, int length, u_char **dnptrs, u_char **lastdnptr)
{
	 u_char *cp, *dn;
	 int c, l;
	u_char **cpp, **lpp, *sp, *eob;
	u_char *msg;

	dn = (u_char *)exp_dn;
	cp = comp_dn;
	eob = cp + length;
	if (dnptrs != NULL) {
		if ((msg = *dnptrs++) != NULL) {
			for (cpp = dnptrs; *cpp != NULL; cpp++)
				;
			lpp = cpp;	/* end of list to search */
		}
	} else
		msg = NULL;
	for (c = *dn++; c != '\0'; ) {
		/* look to see if we can use pointers */
		if (msg != NULL) {
			if ((l = dn_find(dn-1, msg, dnptrs, lpp)) >= 0) {
				if (cp+1 >= eob)
					return (-1);
				*cp++ = (l >> 8) | INDIR_MASK;
				*cp++ = l % 256;
				return (cp - comp_dn);
			}
			/* not found, save it */
			if (lastdnptr != NULL && cpp < lastdnptr-1) {
				*cpp++ = cp;
				*cpp = NULL;
			}
		}
		sp = cp++;	/* save ptr to length byte */
		do {
			if (c == '.') {
				c = *dn++;
				break;
			}
			if (c == '\\') {
				if ((c = *dn++) == '\0')
					break;
			}
			if (cp >= eob) {
				if (msg != NULL)
					*lpp = NULL;
				return (-1);
			}
			*cp++ = c;
		} while ((c = *dn++) != '\0');
		/* catch trailing '.'s but not '..' */
		if ((l = cp - sp - 1) == 0 && c == '\0') {
			cp--;
			break;
		}
		if (l <= 0 || l > MAXLABEL) {
			if (msg != NULL)
				*lpp = NULL;
			return (-1);
		}
		*sp = l;
	}
	if (cp >= eob) {
		if (msg != NULL)
			*lpp = NULL;
		return (-1);
	}
	*cp++ = '\0';
	return (cp - comp_dn);
}

// -------------------------------------------------------------------------

/*
**	Amiga specific stuff
*/

/*
**	Format of host entries
*/

#define ENTRYTIMEOUT (3600*24*90)	// 90 Days timeout

struct nsdbentry {
	time_t read;		// When entry was created
	ULONG ip;
	ULONG hash;
	UBYTE name[244 ];
};

//
// We do some hashing to avoid numerous stricmp()s
//
static ULONG hash( char *name )
{
	ULONG val = strlen( name );

	while( *name )
		val = ( val * 13 ) + ( ToUpper( *name++ ) );

	return( val );
}

static ULONG findhost( char *name )
{
	BPTR fh = Open( "AmiTCP:db/nameserver.db", MODE_OLDFILE );
	ULONG myhash = hash( name );
	struct nsdbentry ns;
	time_t limit = time( NULL ) - ENTRYTIMEOUT;
	int toold = FALSE;
	struct hostent *host;

	// First, look into table
	if( fh )
	{
		SetVBuf( fh, NULL, BUF_FULL, 16384 );
		Flush( fh );

		while( FRead( fh, &ns, sizeof( ns ), 1 ) == 1 )
		{
			// Found!
			if( ns.hash == myhash && !stricmp( ns.name, name ) )
			{
				if( ns.read >= limit )
				{
					// ok!
					Close( fh );
					return( ns.ip );
				}
				// Found, but too old
				toold = TRUE;
				break;
			}
		}
	}

	// We need to do a real lookup
	host = gethostbyname( name );
	if( !host )
	{
		if( fh )
			Close( fh );
		return( ~0 );
	}

	// Create entry
	time( &ns.read );
	memcpy( &ns.ip, &host->h_addr_list[0 ][0 ], 4 );
	ns.hash = myhash;
	strncpy( ns.name, name, 243 );	// This really sucks...
	ns.name[243 ]= 0;

	if( !fh )
		fh = Open( "AmiTCP:db/nameserver.db", MODE_NEWFILE );
	if( fh )
	{
		Flush( fh );
		if( toold )
			Seek( fh, -sizeof ns , OFFSET_CURRENT );
		else
			Seek( fh, 0, OFFSET_END );
		Write( fh, &ns, sizeof ns );
		Close( fh );
	}

	return( ns.ip );
}

static void send_err( int s, char *msg, int msglen, struct sockaddr_in *to, LONG tolen, short rcode )
{
	HEADER *h;
	
	h = ( HEADER * ) msg + sizeof( HEADER );
	h->qr = 1;
	h->rcode = rcode;
	sendto( s, msg, msglen, 0, ( struct sockaddr * ) to, tolen );

}

void __stdargs __main( char *dummy )
{
	int s;						/* Datagram socket */
	struct sockaddr_in from;	/* Address of requestor */
	int fromlen;				/* Length of his address */
	int msglen;					/* Length of the message received */
	ULONG ip;
	UBYTE msg[512 ], *cp, *eom, dnbuf[512 ], *lmsg;
	HEADER *h;
	int n, type, class;
	ULONG xxx = 100000;

	s = init_inet_daemon();
	if( s < 0 )
		exit( 20 );

	SetProgramName( "NameD" );

	openlog( "NameD", LOG_PID | LOG_CONS, LOG_DAEMON );
 
	fromlen = sizeof( from );	

	/* Main loop */
	for(;;)
    {
		if( CheckSignal( SIGBREAKF_CTRL_C ) )
			break;

		/* Read a message */
		fromlen = sizeof(from);
		if( ( msglen = recvfrom( s, msg, sizeof( msg ), 0, (struct sockaddr *)&from, &fromlen ) ) == -1 )
		{
			syslog( LOG_ERR, "recvfrom() error %ld", Errno() );
			continue;
		}
		h = ( HEADER * ) msg;

		if( h->opcode != QUERY )
		{
			send_err( s, msg, msglen, &from, fromlen, NOTIMP );
			continue;
		}

		cp = msg + sizeof( HEADER );
		eom = msg + msglen;

		// Only simple requests
		if( h->qdcount != 1 || h->ancount || h->nscount || h->arcount )
		{
			h->qdcount = 0;
			h->ancount = 0;
			h->nscount = 0;
			h->arcount = 0;
			send_err( s, msg, msglen, &from, fromlen, FORMERR );
			continue;
		}
		if( ( n = dn_expand( msg, eom, cp, dnbuf, sizeof( dnbuf ) ) ) < 0 )
		{
			send_err(s,msg,msglen,&from,fromlen,FORMERR);
			continue;
		}
		cp += n;
		GETSHORT( type, cp );
		GETSHORT( class, cp );
		if( cp > eom )
		{
			send_err( s, msg, msglen, &from, fromlen, FORMERR );
			continue;
		}
		if( type != T_A || class != C_IN )
		{
			send_err(s,msg,msglen,&from,fromlen,NOTIMP);
			continue;
		}

		// find IP address of host
		ip = findhost( dnbuf );
			
		if( ip == ~0 )
		{
			send_err( s, msg, msglen, &from, fromlen, NXDOMAIN );
			continue;
		}
		
		lmsg = msg + sizeof( msg );
		if( ( n = dn_comp( dnbuf, cp, lmsg - cp, NULL, NULL ) ) < 0 )
		{
			send_err( s, msg, msglen, &from, fromlen, SERVFAIL );
			continue;						
		}
		cp += n;
		if( cp + 14 > lmsg )
		{
			send_err( s, msg, msglen, &from, fromlen, SERVFAIL );
			continue;			
		}
		
		PUTSHORT( type, cp );
		PUTSHORT( class, cp );
		PUTLONG( xxx, cp );
		PUTSHORT( sizeof( LONG ), cp );
		PUTLONG( ip, cp );
		
		h->qr = 1;
		h->rcode = NOERROR;
		h->ancount = 1;
		sendto( s, msg, cp - msg, 0, ( struct sockaddr * ) &from, fromlen );
    }
}

// EOF
